/*
 * Snes9x - Portable Super Nintendo Entertainment System (TM) emulator.
 *
 * (c) Copyright 1996 - 2001 Gary Henderson (gary.henderson@ntlworld.com) and
 *                           Jerremy Koot (jkoot@snes9x.com)
 *
 * Super FX C emulator code 
 * (c) Copyright 1997 - 1999 Ivar (ivar@snes9x.com) and
 *                           Gary Henderson.
 * Super FX assembler emulator code (c) Copyright 1998 zsKnight and _Demo_.
 *
 * DSP1 emulator code (c) Copyright 1998 Ivar, _Demo_ and Gary Henderson.
 * C4 asm and some C emulation code (c) Copyright 2000 zsKnight and _Demo_.
 * C4 C code (c) Copyright 2001 Gary Henderson (gary.henderson@ntlworld.com).
 *
 * DOS port code contains the works of other authors. See headers in
 * individual files.
 *
 * Snes9x homepage: http://www.snes9x.com
 *
 * Permission to use, copy, modify and distribute Snes9x in both binary and
 * source form, for non-commercial purposes, is hereby granted without fee,
 * providing that this license information and copyright notice appear with
 * all copies and any derived work.
 *
 * This software is provided 'as-is', without any express or implied
 * warranty. In no event shall the authors be held liable for any damages
 * arising from the use of this software.
 *
 * Snes9x is freeware for PERSONAL USE only. Commercial users should
 * seek permission of the copyright holders first. Commercial use includes
 * charging money for Snes9x or software derived from Snes9x.
 *
 * The copyright holders request that bug fixes and improvements to the code
 * should be forwarded to them so everyone can benefit from the modifications
 * in future versions.
 *
 * Super NES and Super Nintendo Entertainment System are trademarks of
 * Nintendo Co., Limited and its subsidiary companies.
 */

#ifdef DREAMCAST
#include <kos.h>
extern uint8 romdisk[];
KOS_INIT_FLAGS(INIT_DEFAULT);
KOS_INIT_ROMDISK(romdisk);
#endif


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <SDL.h>
#include <SDL_mixer.h>
#ifdef DREAMCAST
#include <SDL_dreamcast.h>
#endif
#include <signal.h>
#include <errno.h>

#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/types.h>
#include <ctype.h>
#include <dirent.h>
#include "keydef.h"

#include "snes9x.h"
#include "memmap.h"
#include "debug.h"
#include "cpuexec.h"
#include "ppu.h"
#include "snapshot.h"
#include "apu.h"
#include "display.h"
#include "gfx.h"
#include "soundux.h"
#include "spc700.h"
#include "sa1.h"

#include "menu.h"

#include "sdlinterface.h"

#define AUTO_FRAMESKIP_DELAY 10
#define TH_VCOUNTER_SPLIT 11
#define SNES_MAX_MID_VCOUNTER (SNES_MAX_NTSC_VCOUNTER+((SNES_MAX_PAL_VCOUNTER-SNES_MAX_NTSC_VCOUNTER)/2))
#define TH_VCOUNTER0 ((TH_VCOUNTER_SPLIT*SNES_MAX_PAL_VCOUNTER)/10)
#define TH_VCOUNTER1 TH_VCOUNTER0
#define TH_VCOUNTER2 ((TH_VCOUNTER_SPLIT*SNES_MAX_MID_VCOUNTER)/10)
#define TH_VCOUNTER3 ((TH_VCOUNTER_SPLIT*SNES_MAX_NTSC_VCOUNTER)/10)
#define TH_VCOUNTER4 ((TH_VCOUNTER_SPLIT*240)/10)


unsigned snes4all_max_vcounter=TH_VCOUNTER0;
unsigned snes4all_apu_hmax=1;
long snes4all_vcounter[512];

#if defined(DEBUG_TILECACHE) || defined(PROFILER_SNES4ALL)
int snes4all_frameskip=0;
int snes4all_sound_enable=0;
int snes4all_transparency=0;
int snes4all_throttle=0;
int snes4all_sound_throttle=0;
unsigned snes4all_timeslice=2;

#ifdef DEBUG_TILECACHE
unsigned snes4all_debug_tilecache_tiles=0;
unsigned snes4all_debug_tilecache_fails=0;

static void print_tileche_eficience(void)
{
	double d=snes4all_debug_tilecache_fails;
	d*=100.0;
	d/=(double)snes4all_debug_tilecache_tiles;
	d=100.0-d;
	printf("%i total, %i fallados, EFICIENCIA %.2f%%\n",snes4all_debug_tilecache_tiles,snes4all_debug_tilecache_fails,d);
}
#endif

#else
int snes4all_frameskip=-1;
int snes4all_sound_enable=-1;
int snes4all_transparency=0;
int snes4all_throttle=3;
int snes4all_sound_throttle=((1>>3)-1);
unsigned snes4all_timeslice=2+(3<<1);
#endif

int OldSkipFrame;

extern SDL_Surface *screen;

long reportf=0;

#define MUSIC_VOLUME 64

#define MAX_FRAMESKIP 10
 
char snes4all_image_file[1024];
int snes4all_emulating=0;
char *rom_filename = (char *)&snes4all_image_file[0];
char *snapshot_filename = NULL;


static uint32 joypad_val[8]={0x80000000,0x80000000,0x80000000,0x80000000,0x80000000,0x80000000,0x80000000,0x80000000};

void OutOfMemory ()
{
    fprintf (stderr, "\
Snes9X: Memory allocation failure - not enough RAM/virtual memory available.\n\
        S9xExiting...\n");
    Memory.Deinit ();
    S9xDeinitAPU ();
    
    exit (1);
}


static void getChanges(void)
{
	if (snes4all_transparency)
		Settings.Transparency=TRUE;
	else
		Settings.Transparency=FALSE;
	if (snes4all_frameskip<0)
		Settings.SkipFrames=AUTO_FRAMERATE;
	else
		Settings.SkipFrames=snes4all_frameskip;
#ifdef NOSOUND
	S9xSetSoundMute(!snes4all_sound_enable);
#endif

	long i;
	switch(snes4all_throttle)
	{
		case 4:
			Settings.CyclesPercentage = (100 * SNES_MAX_PAL_VCOUNTER) / TH_VCOUNTER4;
			for(i=0;i<240;i++)
				snes4all_vcounter[i]=i+1;
			for(i=240;i<512;i++)
				snes4all_vcounter[i]=511;
			break;
		case 3:
			Settings.CyclesPercentage = (100 * SNES_MAX_PAL_VCOUNTER) / TH_VCOUNTER3;
			for(i=0;i<SNES_MAX_NTSC_VCOUNTER;i++)
				snes4all_vcounter[i]=i+1;
			for(i=SNES_MAX_NTSC_VCOUNTER;i<512;i++)
				snes4all_vcounter[i]=511;
			break;
		case 2:
			Settings.CyclesPercentage = (100 * SNES_MAX_PAL_VCOUNTER) / TH_VCOUNTER2;
			for(i=0;i<SNES_MAX_MID_VCOUNTER;i++)
				snes4all_vcounter[i]=i+1;
			for(i=SNES_MAX_MID_VCOUNTER;i<512;i++)
				snes4all_vcounter[i]=511;
			break;
		case 1:
			Settings.CyclesPercentage = (100 * SNES_MAX_PAL_VCOUNTER) / TH_VCOUNTER1;
			for(i=0;i<SNES_MAX_PAL_VCOUNTER;i++)
				snes4all_vcounter[i]=i+1;
			for(i=SNES_MAX_PAL_VCOUNTER;i<512;i++)
				snes4all_vcounter[i]=511;
			break;
		default:
			Settings.CyclesPercentage = (100 * SNES_MAX_PAL_VCOUNTER) / TH_VCOUNTER0;
			for(i=0;i<512;i++)
				snes4all_vcounter[i]=i+1;
	}
	snes4all_sound_throttle=((1>>snes4all_throttle)-1);
	snes4all_vcounter[511]=511;
	snes4all_timeslice=2+(snes4all_throttle<<1);
	Settings.H_Max = (SNES_CYCLES_PER_SCANLINE * Settings.CyclesPercentage)/100;
	snes4all_apu_hmax=Settings.H_Max*10000L;

	SDL_Event e;
	SDL_Delay(100);
	while (SDL_PollEvent(&e) > 0)
		SDL_Delay(10);
}

#define dCPU(A) ((unsigned)(A))-((unsigned)&CPU)
#define dSA1(A) ((unsigned)(A))-((unsigned)&SA1)
#define dIAPU(A) ((unsigned)(A))-((unsigned)&IAPU)

/*#include "cheats.h"*/
extern "C"
int main (int argc, char **argv)
{
    ZeroMemory (&Settings, sizeof (Settings));

    Settings.C4=FALSE;
    Settings.ForceC4=FALSE;
    Settings.SoundPlaybackRate = 4;
#ifdef SNES4ALL_STEREO
    Settings.Stereo = TRUE;
#else
    Settings.Stereo = FALSE;
#endif
    Settings.SoundBufferSize = 1024;
    Settings.JoystickEnabled = TRUE;
    Settings.APUEnabled = Settings.NextAPUEnabled = TRUE;
    Settings.CyclesPercentage = 100;
    Settings.H_Max = SNES_CYCLES_PER_SCANLINE;
    snes4all_apu_hmax=Settings.H_Max*10000L;
    Settings.SkipFrames = AUTO_FRAMERATE;
    Settings.ShutdownMaster = TRUE;
    Settings.FrameTimePAL = 20;
    Settings.FrameTimeNTSC = 17;
    Settings.FrameTime = Settings.FrameTimePAL;
    Settings.Mouse = FALSE;
    Settings.SuperScope = FALSE;
    Settings.MultiPlayer5 = FALSE;
//    Settings.ControllerOption = SNES_MULTIPLAYER5;
    Settings.ControllerOption = 0;
    Settings.Transparency = FALSE; //TRUE;
#ifdef NETPLAY_SUPPORT
    Settings.NetPlay = FALSE;
    Settings.ServerName [0] = 0;
#endif
    Settings.AutoSaveDelay = 30;
    Settings.ApplyCheats = TRUE;
    Settings.TurboMode = FALSE;
    Settings.TurboSkipFrames = 15;


    Settings.HBlankStart = (256 * Settings.H_Max) / SNES_HCOUNTER_MAX;

    if (!Memory.Init () || !S9xInitAPU())
	OutOfMemory ();

    uint32 saved_flags = CPU.Flags;

#ifdef GFX_MULTI_FORMAT
    S9xSetRenderPixelFormat (RGB565);
#endif

puts("MAIN!!!!");
    S9xInitDisplay (0, NULL);
    if (!S9xGraphicsInit ())
	OutOfMemory ();

   (void) S9xInitSound (Settings.SoundPlaybackRate, Settings.Stereo,
			 Settings.SoundBufferSize);

#ifdef NOSOUND
    Settings.APUEnabled=FALSE;
#else
    if (!Settings.APUEnabled)
#endif
	S9xSetSoundMute (TRUE);

    S9xInitInputDevices ();

    snes4all_image_file[0]=0;
    init_text(1);
#ifndef AUTOLOAD
    run_mainMenu();
#else
    strcpy(rom_filename,AUTOLOAD);
#endif
    SDL_FillRect(screen,NULL,0);
    sound_disable_music();
    quit_text();
    snes4all_emulating=1;

    getChanges();

    if (rom_filename)
    {
	if (!Memory.LoadROM (rom_filename))
	{
	    char dir [_MAX_DIR + 1];
	    char drive [_MAX_DRIVE + 1];
	    char name [_MAX_FNAME + 1];
	    char ext [_MAX_EXT + 1];
	    char fname [_MAX_PATH + 1];

	    _splitpath (rom_filename, drive, dir, name, ext);
	    _makepath (fname, drive, dir, name, ext);

	    strcpy (fname, S9xGetROMDirectory ());
	    strcat (fname, SLASH_STR);
	    strcat (fname, name);
	    if (ext [0])
	    {
		strcat (fname, ".");
		strcat (fname, ext);
	    }
	    _splitpath (fname, drive, dir, name, ext);
	    _makepath (fname, drive, dir, name, ext);
	    if (!Memory.LoadROM (fname))
	    {
		printf ("Error opening: %s\n", rom_filename);
		exit (1);
	    }
	}
	Memory.LoadSRAM (S9xGetFilename (".srm"));
//	S9xLoadCheatFile (S9xGetFilename (".cht"));
    }
    else
    {
	S9xReset ();
	Settings.Paused |= 2;
    }
    CPU.Flags = saved_flags;

    if (snapshot_filename)
    {
	int Flags = CPU.Flags & (DEBUG_MODE_FLAG | TRACE_FLAG);
	if (!S9xLoadSnapshot (snapshot_filename))
	    exit (1);
	CPU.Flags |= Flags;
    }
#ifndef _ZAURUS
    S9xGraphicsMode ();
    sprintf (String, "\"%s\" %s: %s", Memory.ROMName, TITLE, VERSION);
    S9xSetTitle (String);
#endif

   S9xSetSoundMute (TRUE);
   if (!Settings.APUEnabled)
	S9xSetSoundMute (FALSE);

#ifdef PROFILER_SNES4ALL
   snes4all_prof_init();
   snes4all_prof_add("S9xMainLoop");			// 0
   snes4all_prof_add("S9xDoHBlankProcessing");		// 1
   snes4all_prof_add("RenderLine");			// 2
   snes4all_prof_add("S9xEndScreenRefresh");		// 3
#endif

   snes4all_prof_start(0);

#ifndef DEBUGGER
   S9xMainLoop ();
#else
    while (1)
    {
	if (!Settings.Paused
	    || (CPU.Flags & (DEBUG_MODE_FLAG | SINGLE_STEP_FLAG))
           )

	    S9xMainLoop ();

	if (Settings.Paused
	    || (CPU.Flags & DEBUG_MODE_FLAG)
           )
	{
	    S9xSetSoundMute (TRUE);
	}

	if (CPU.Flags & DEBUG_MODE_FLAG)
	{
	    S9xDoDebug ();
	}
	else
	if (Settings.Paused)
	    S9xProcessEvents (TRUE);

	if (!Settings.Paused
	    && !(CPU.Flags & DEBUG_MODE_FLAG)
           )
	{
	    S9xSetSoundMute (FALSE);
	}
    }
#endif
    return (0);
}

void S9xAutoSaveSRAM ()
{
    Memory.SaveSRAM (S9xGetFilename (".srm"));
}

void S9xExit ()
{
    S9xSetSoundMute (TRUE);
    S9xDeinitDisplay ();
    Memory.SaveSRAM (S9xGetFilename (".srm"));
//    S9xSaveCheatFile (S9xGetFilename (".cht"));
    Memory.Deinit ();
    S9xDeinitAPU ();

    exit (0);
}

void S9xInitInputDevices ()
{
}

const char *GetHomeDirectory ()
{
    return (getenv ("HOME"));
}

const char *S9xGetSnapshotDirectory ()
{
    static char filename [PATH_MAX];
    const char *snapshot;
    
    if (!(snapshot = getenv ("SNES9X_SNAPSHOT_DIR")) &&
	!(snapshot = getenv ("SNES96_SNAPSHOT_DIR")))
    {
#ifdef WIN32
	return("/tmp");
#else
#ifdef DREAMCAST
	return("/ram");
#else
	const char *home = GetHomeDirectory ();
	strcpy (filename, home);
	strcat (filename, SLASH_STR);
	strcat (filename, ".snes96_snapshots");
	mkdir (filename, 0777);
	chown (filename, getuid (), getgid ());
#endif
#endif
    }
    else
	return (snapshot);

    return (filename);
}

const char *S9xGetFilename (const char *ex)
{
    static char filename [PATH_MAX + 1];
    char drive [_MAX_DRIVE + 1];
    char dir [_MAX_DIR + 1];
    char fname [_MAX_FNAME + 1];
    char ext [_MAX_EXT + 1];

    _splitpath (Memory.ROMFilename, drive, dir, fname, ext);
    strcpy (filename, S9xGetSnapshotDirectory ());
    strcat (filename, SLASH_STR);
    strcat (filename, fname);
    strcat (filename, ex);

    return (filename);
}

const char *S9xGetROMDirectory ()
{
    const char *roms;
    
    if (!(roms = getenv ("SNES9X_ROM_DIR")) &&
	!(roms = getenv ("SNES96_ROM_DIR")))
	return ("." SLASH_STR "roms");
    else
	return (roms);
}

const char *S9xBasename (const char *f)
{
    const char *p;
    if ((p = strrchr (f, '/')) != NULL || (p = strrchr (f, '\\')) != NULL)
	return (p + 1);

    return (f);
}

#ifndef _ZAURUS
const char *S9xChooseFilename (bool8 read_only)
{
    char def [PATH_MAX + 1];
    char title [PATH_MAX + 1];
    char drive [_MAX_DRIVE + 1];
    char dir [_MAX_DIR + 1];
    char ext [_MAX_EXT + 1];

    _splitpath (Memory.ROMFilename, drive, dir, def, ext);
    strcat (def, ".s96");
    sprintf (title, "%s snapshot filename",
	    read_only ? "Select load" : "Choose save");
    const char *filename;

    S9xSetSoundMute (TRUE);
    filename = S9xSelectFilename (def, S9xGetSnapshotDirectory (), "s96", title);
    S9xSetSoundMute (FALSE);
    return (filename);
}
#endif

bool8 S9xOpenSnapshotFile (const char *fname, bool8 read_only, STREAM *file)
{
    char filename [PATH_MAX + 1];
    char drive [_MAX_DRIVE + 1];
    char dir [_MAX_DIR + 1];
    char ext [_MAX_EXT + 1];

    _splitpath (fname, drive, dir, filename, ext);

    if (*drive || *dir == '/' ||
	(*dir == '.' && (*(dir + 1) == '/'
        )))
    {
	strcpy (filename, fname);
	if (!*ext)
	    strcat (filename, ".s96");
    }
    else
    {
	strcpy (filename, S9xGetSnapshotDirectory ());
	strcat (filename, SLASH_STR);
	strcat (filename, fname);
	if (!*ext)
	    strcat (filename, ".s96");
    }
    
#ifdef ZLIB
    if (read_only)
    {
	if ((*file = OPEN_STREAM (filename, "rb")))
	    return (TRUE);
    }
    else
    {
	if ((*file = OPEN_STREAM (filename, "wb")))
	    return (TRUE);
    }
#else
    char command [PATH_MAX];
    
    if (read_only)
    {
	sprintf (command, "gzip -d <\"%s\"", filename);
	if (*file = popen (command, "r"))
	    return (TRUE);
    }
    else
    {
	sprintf (command, "gzip --best >\"%s\"", filename);
	if (*file = popen (command, "wb"))
	    return (TRUE);
    }
#endif
    return (FALSE);
}

void S9xCloseSnapshotFile (STREAM file)
{
#ifdef ZLIB
    CLOSE_STREAM (file);
#else
    pclose (file);
#endif
}

#ifndef _ZAURUS
bool8 S9xInitUpdate ()
{
    return (TRUE);
}
#endif

bool8 S9xDeinitUpdate (int Width, int Height,bool8 sixnocount)
{
        SDL_Flip(screen);
	return(TRUE);
}


// #define now() SDL_GetTicks()

void _makepath (char *path, const char *, const char *dir,
		const char *fname, const char *ext)
{
    if (dir && *dir)
    {
	strcpy (path, dir);
	strcat (path, "/");
    }
    else
	*path = 0;
    strcat (path, fname);
    if (ext && *ext)
    {
        strcat (path, ".");
        strcat (path, ext);
    }
}

void _splitpath (const char *path, char *drive, char *dir, char *fname,
		 char *ext)
{
    *drive = 0;

    char *slash = strrchr (path, '/');
    if (!slash)
	slash = strrchr (path, '\\');

    char *dot = strrchr (path, '.');

    if (dot && slash && dot < slash)
	dot = NULL;

    if (!slash)
    {
	strcpy (dir, "");
	strcpy (fname, path);
        if (dot)
        {
	    *(fname + (dot - path)) = 0;
	    strcpy (ext, dot + 1);
        }
	else
	    strcpy (ext, "");
    }
    else
    {
	strcpy (dir, path);
	*(dir + (slash - path)) = 0;
	strcpy (fname, slash + 1);
        if (dot)
	{
	    *(fname + (dot - slash) - 1) = 0;
    	    strcpy (ext, dot + 1);
	}
	else
	    strcpy (ext, "");
    }
}

/*
#ifndef _ZAURUS
void S9xToggleSoundChannel (int c)
{
    if (c == 8)
	so.sound_switch = 255;
    else
	so.sound_switch ^= 1 << c;
    S9xSetSoundControl (so.sound_switch);
}
#endif
*/

#ifdef DEBUG_FRAMESKIP
static unsigned snes4all_fskip_nframes=0;
static unsigned snes4all_fskip_skipped=0;
static double snes4all_fskip_framerate=0.0;

static void print_frameskip(void)
{
	double d=(((double)snes4all_fskip_skipped)*100.0)/((double)snes4all_fskip_nframes);
	printf("%i frames, skipped %i (%.2f%%)\n",snes4all_fskip_nframes,snes4all_fskip_skipped,d);
	printf("Framerate %.4f\n", snes4all_fskip_framerate);
}

static void debug_frameskip(bool isframe)
{
	static Uint32 start_time=0;
	static Uint32 start_numframes=0;

	if (!isframe)
		snes4all_fskip_skipped++;
	snes4all_fskip_nframes++;
	if (!start_time)
	{
		start_time=SDL_GetTicks();
		start_numframes=snes4all_fskip_nframes;
	}
	else
	{
		Uint32 now=SDL_GetTicks();
		if (now-start_time>=1000)
		{
			if (snes4all_fskip_framerate!=0.0)
				snes4all_fskip_framerate=(snes4all_fskip_framerate+((double)(snes4all_fskip_nframes-start_numframes)))/2.0;
			else
				snes4all_fskip_framerate=(double)(snes4all_fskip_nframes-start_numframes);
			start_time=now;
			start_numframes=snes4all_fskip_nframes;
		}
	}
#ifdef AUTO_DEBUG_FRAMESKIP
	if (snes4all_fskip_nframes>=AUTO_DEBUG_FRAMESKIP)
	{
		print_frameskip();
		exit(0);
	}
#endif

}

#endif

void S9xSyncSpeed ()
{
    S9xProcessEvents (FALSE);
#if defined(AUTO_DEBUG_TILECACHE) && defined(DEBUG_TILECACHE)
    {
	    static unsigned cu=0;
	    if (cu>=AUTO_DEBUG_TILECACHE)
	    {
		    print_tileche_eficience();
		    exit(0);
	    }
	    cu++;
    }
#endif
    static Uint32 next1=0;
//    if (!Settings.TurboMode && Settings.SkipFrames == AUTO_FRAMERATE)
    if (Settings.SkipFrames == AUTO_FRAMERATE)
    {
	Uint32 now=SDL_GetTicks();

	if (!next1)
	    next1 = now;

	next1 += Settings.FrameTime;

	if ((now-AUTO_FRAMESKIP_DELAY)>next1)
	{
	    IPPU.SkippedFrames++;
	    if (IPPU.SkippedFrames > MAX_FRAMESKIP)
	    {
		IPPU.RenderThisFrame = TRUE;
		IPPU.SkippedFrames = 0;
		next1 = now + 2;
	    }
	    else
		IPPU.RenderThisFrame = FALSE;
	}
	else
	{
	    CHECK_SOUND ();
	    if ((now+AUTO_FRAMESKIP_DELAY)<next1)
	    {
#ifndef DREAMCAST
		do
		{
		    CHECK_SOUND ();
		    SDL_Delay(1);
//		    S9xProcessEvents (FALSE);
		    now=SDL_GetTicks();
		} while (next1 > now);
#endif
		next1=now;
	    }
	    IPPU.RenderThisFrame = TRUE;
	    IPPU.SkippedFrames = 0;
	}
    }
    else
    {
	next1 = 0;
	if (++IPPU.FrameSkip >= (Settings.TurboMode ? Settings.TurboSkipFrames
						    : Settings.SkipFrames))
	{
	    IPPU.FrameSkip = 0;
	    IPPU.SkippedFrames = 0;
	    IPPU.RenderThisFrame = TRUE;
	}
	else
	{
	    IPPU.SkippedFrames++;
	    IPPU.RenderThisFrame = FALSE;
	}
    }
#ifdef DEBUG_FRAMESKIP
    debug_frameskip(IPPU.RenderThisFrame);
#endif
}


#ifndef AUTO_EVENTS

static void goMenu(void)
{
	int eret;
	snes4all_emulating=1;
#ifdef DEBUG_TILECACHE
	print_tileche_eficience();
#endif
#ifdef DEBUG_FRAMESKIP
	print_frameskip();
#endif
	rom_filename[0]=0;
    	sound_enable_music();
	init_text(0);
    	eret=run_mainMenu();
	quit_text();
    	sound_disable_music();
	if(eret==2)
		S9xReset();
	getChanges();
	if (rom_filename[0])
		Memory.LoadROM (rom_filename);
	SDL_FillRect(screen,NULL,0);
}

static void keyprocess(SDLKey key, SDL_bool pressed)
{
	uint32 val=0,which=0;

	switch(key)
	{
		case SDLK_RETURN:
			which=0;
			val=SNES_START_MASK;
			break;
		case SDLK_LCTRL:
			which=0;
			val=SNES_B_MASK;
			break;
		case SDLK_LALT:
			which=0;
			val=SNES_A_MASK;
			break;
		case SDLK_SPACE:
			which=0;
			val=SNES_X_MASK;
			break;
		case SDLK_LSHIFT:
			which=0;
			val=SNES_Y_MASK;
			break;
		case SDLK_TAB:
			which=0;
			val=SNES_TR_MASK;
			break;
		case SDLK_BACKSPACE:
			which=0;
			val=SNES_TL_MASK;
			break;
		case SDLK_LEFT:
			which=0;
			val=SNES_LEFT_MASK;
			break;
		case SDLK_RIGHT:
			which=0;
			val=SNES_RIGHT_MASK;
			break;
		case SDLK_UP:
			which=0;
			val=SNES_UP_MASK;
			break;
		case SDLK_DOWN:
			which=0;
			val=SNES_DOWN_MASK;
			break;
#ifndef DREAMCAST
		case SDLK_BACKSLASH:
			which=0;
			val=SNES_SELECT_MASK;
			break;
#endif


		case SDLK_z:
			which=1;
			val=SNES_START_MASK;
			break;
		case SDLK_e:
			which=1;
			val=SNES_B_MASK;
			break;
		case SDLK_q:
			which=1;
			val=SNES_A_MASK;
			break;
		case SDLK_x:
			which=1;
			val=SNES_X_MASK;
			break;
		case SDLK_c:
			which=1;
			val=SNES_Y_MASK;
			break;
		case SDLK_1:
			which=1;
			val=SNES_TR_MASK;
			break;
		case SDLK_2:
			which=1;
			val=SNES_TL_MASK;
			break;
		case SDLK_a:
			which=1;
			val=SNES_LEFT_MASK;
			break;
		case SDLK_d:
			which=1;
			val=SNES_RIGHT_MASK;
			break;
		case SDLK_w:
			which=1;
			val=SNES_UP_MASK;
			break;
		case SDLK_s:
			which=1;
			val=SNES_DOWN_MASK;
			break;
#ifndef DREAMCAST
		case SDLK_3:
			which=1;
			val=SNES_SELECT_MASK;
			break;
#endif



		case SDLK_v:
			which=2;
			val=SNES_START_MASK;
			break;
		case SDLK_y:
			which=2;
			val=SNES_B_MASK;
			break;
		case SDLK_r:
			which=2;
			val=SNES_A_MASK;
			break;
		case SDLK_b:
			which=2;
			val=SNES_X_MASK;
			break;
		case SDLK_n:
			which=2;
			val=SNES_Y_MASK;
			break;
		case SDLK_4:
			which=2;
			val=SNES_TR_MASK;
			break;
		case SDLK_5:
			which=2;
			val=SNES_TL_MASK;
			break;
		case SDLK_f:
			which=2;
			val=SNES_LEFT_MASK;
			break;
		case SDLK_h:
			which=2;
			val=SNES_RIGHT_MASK;
			break;
		case SDLK_t:
			which=2;
			val=SNES_UP_MASK;
			break;
		case SDLK_g:
			which=2;
			val=SNES_DOWN_MASK;
			break;
#ifndef DREAMCAST
		case SDLK_7:
			which=2;
			val=SNES_SELECT_MASK;
			break;
#endif



		case SDLK_m:
			which=3;
			val=SNES_START_MASK;
			break;
		case SDLK_o:
			which=3;
			val=SNES_B_MASK;
			break;
		case SDLK_u:
			which=3;
			val=SNES_A_MASK;
			break;
		case SDLK_COMMA:
			which=3;
			val=SNES_X_MASK;
			break;
		case SDLK_PERIOD:
			which=3;
			val=SNES_Y_MASK;
			break;
		case SDLK_8:
			which=3;
			val=SNES_TR_MASK;
			break;
		case SDLK_9:
			which=3;
			val=SNES_TL_MASK;
			break;
		case SDLK_j:
			which=3;
			val=SNES_LEFT_MASK;
			break;
		case SDLK_l:
			which=3;
			val=SNES_RIGHT_MASK;
			break;
		case SDLK_i:
			which=3;
			val=SNES_UP_MASK;
			break;
		case SDLK_k:
			which=3;
			val=SNES_DOWN_MASK;
			break;
#ifndef DREAMCAST
		case SDLK_0:
			which=3;
			val=SNES_SELECT_MASK;
			break;
#endif

		default:
			return;
	}

	if (pressed)
		joypad_val[which]|=val;
	else
		joypad_val[which]^=val;

//	printf("%s %s -> JOYPAD[%i]=0x%X\n",pressed?"Press":"Unpress", SDL_GetKeyName(key),which,joypad_val[which]);
}
#endif

void S9xProcessEvents (bool8 block)
{
#ifdef AUTO_EVENTS
	static unsigned cuenta=0;

	switch(cuenta&127)
	{
		case 3:
			joypad_val[0]|=SNES_START_MASK;
			break;
		case 10:
			joypad_val[0]^=SNES_START_MASK;
			break;
	}
#ifdef MAX_AUTO_EVENTS
	if (cuenta>MAX_AUTO_EVENTS)
	{
#ifdef PROFILER_SNES4ALL
		snes4all_prof_show();
#endif
#ifdef DEBUG_FRAMESKIP
		print_frameskip();
#endif
#ifdef DEBUG_TILECACHE
		print_tileche_eficience();
#endif
		exit(0);
	}
#endif
	cuenta++;
#else
#ifndef DREAMCAST
	uint32 num = 0;
	static bool8 TURBO = FALSE;
#endif

	SDL_Event event;
	while(SDL_PollEvent(&event)) {
		switch(event.type) {
		case SDL_QUIT:
			S9xExit();
			break;
#ifdef DREAMCAST
		case SDL_JOYAXISMOTION:
			if (event.jaxis.axis==0)
			{
				if (event.jaxis.value<-100)
					goMenu();
			}
			else
			if (event.jaxis.axis==1)
			{
				if (event.jaxis.value>100)
					joypad_val[event.jaxis.which]|=SNES_SELECT_MASK;
				else 
					joypad_val[event.jaxis.which]^=SNES_SELECT_MASK;
			}
			break;
#endif
		case SDL_KEYDOWN:
			keyprocess(event.key.keysym.sym,SDL_TRUE);

#ifndef DREAMCAST
			if (event.key.keysym.sym == SDLK_ESCAPE)
				S9xExit();
			else
			if (event.key.keysym.sym == SDLK_F11)
				goMenu();
			else
			if (event.key.keysym.sym == SDLK_F12)
				SDL_WM_ToggleFullScreen(screen);
			else
			if (event.key.keysym.sym == SDLK_PAGEUP)
			{
				if (!TURBO) {
					TURBO = TRUE;
					OldSkipFrame = Settings.SkipFrames;
					Settings.SkipFrames = 10;
				}
			}
			else
		if (event.key.keysym.sym == SDLK_F1)	PPU.BG_Forced ^= 1;
		    else if (event.key.keysym.sym == SDLK_F2)	PPU.BG_Forced ^= 2;
		    else if (event.key.keysym.sym == SDLK_F3)	PPU.BG_Forced ^= 4;
		    else if (event.key.keysym.sym == SDLK_F4)	PPU.BG_Forced ^= 8;
		    else if (event.key.keysym.sym == SDLK_F5)	PPU.BG_Forced ^= 16;
			else if (event.key.keysym.sym == SDLK_F6)	num = 1;
			else if (event.key.keysym.sym == SDLK_F7)	num = 2;
			else if (event.key.keysym.sym == SDLK_F8)	num = 3;
			else if (event.key.keysym.sym == SDLK_F9)	num = 4;
			else if (event.key.keysym.sym == SDLK_F10) {
					S9xReset();
			}
			if (num) {
				char fname[256], ext[8];
				sprintf(ext, ".00%d", num - 1);
				strcpy(fname, S9xGetFilename (ext));
			    if (event.key.keysym.mod & KMOD_SHIFT)
				    S9xFreezeGame (fname);
				else
					S9xLoadSnapshot (fname);
			}
#endif
			break;
		case SDL_KEYUP:
			keyprocess(event.key.keysym.sym,SDL_FALSE);
#ifndef DREAMCAST
			if (event.key.keysym.sym == SDLK_PAGEUP)
				if (TURBO) {
					TURBO = FALSE;
					Settings.SkipFrames = OldSkipFrame;
				}
#endif
			break;
		}
	}
#endif
}

#ifdef _ZAURUS
static long log2 (long num)
{
    long n = 0;

    while (num >>= 1)
	n++;

    return (n);
}
#endif

static int Rates[8] =
{
    0, 8192, 11025, 16500, 22050, 29300, 36600, 44000
};

static int BufferSizes [8] =
{
    0, 256, 256, 256, 512, 512, 1024, 1024
};

static Mix_Chunk *sample_beep=NULL;
static Mix_Music *menu_music=NULL;

void update_sdl_audio(void *userdata, Uint8 * stream, int len);
void sound_play_menu_music(void)
{
    if (menu_music)
    {
    	Mix_PlayMusic(menu_music,-1);
    	Mix_VolumeMusic(MUSIC_VOLUME);
    }
}

void sound_play_beep(void) 
{
	if (sample_beep)
		Mix_PlayChannel(0,sample_beep,0);
}

void sound_enable_music(void)
{
	SDL_PauseAudio(1);
#ifdef DREAMCAST
	SDL_DC_RestoreSoundBuffer();
#endif
	Mix_HookMusic(NULL,NULL);
	SDL_PauseAudio(0);
	Mix_VolumeMusic(MUSIC_VOLUME);
}

void sound_disable_music(void)
{
	SDL_PauseAudio(1);
	Mix_HookMusic(&update_sdl_audio,NULL);
	SDL_PauseAudio(0);
}


bool8 S9xOpenSoundDevice (int mode, bool8 stereo, int buffer_size)
{
    so.stereo = stereo;
    so.playback_rate = Rates[mode & 0x07];
    S9xSetPlaybackRate (so.playback_rate);

    if (buffer_size == 0)
	buffer_size = BufferSizes [mode & 7];
    if (buffer_size > MAX_BUFFER_SIZE / 4)
	buffer_size = MAX_BUFFER_SIZE / 4;
    so.buffer_size=buffer_size;
    Mix_OpenAudio(so.playback_rate,AUDIO_S16,so.stereo ? 2:1,so.buffer_size);
//printf("Mix_OpenAudio(%i,AUDIO_S16,%i,%i);\n",so.playback_rate,so.stereo ? 2:1,so.buffer_size);
    sample_beep=Mix_LoadWAV(DATA_PREFIX "beep.wav");
    menu_music=Mix_LoadMUS(DATA_PREFIX "music.mod");
    return (TRUE);
}

static uint8 Buf[MAX_BUFFER_SIZE];

#define FIXED_POINT 0x10000
#define FIXED_POINT_SHIFT 16
#define FIXED_POINT_REMAINDER 0xffff

void S9xGenerateSound ()
{
#ifndef NOSOUND
    if (!snes4all_sound_enable)
	    return;
    int buffer_size = so.buffer_size;
    int bytes_so_far = (so.samples_mixed_so_far << 1);
    if (bytes_so_far >= buffer_size)
	return;


    so.err_counter += so.err_rate;
    if (so.err_counter >= FIXED_POINT)
    {
        int sample_count = so.err_counter >> FIXED_POINT_SHIFT;
	int byte_offset;

        so.err_counter &= FIXED_POINT_REMAINDER;
	if (so.stereo)
	    sample_count <<= 1;
	byte_offset = bytes_so_far + so.play_position;
	    
	do
	{
	    int sc = sample_count;
	    int byte_count = sample_count;
	    byte_count <<= 1;
	    
	    if ((byte_offset & SOUND_BUFFER_SIZE_MASK) + byte_count > SOUND_BUFFER_SIZE)
	    {
		sc = SOUND_BUFFER_SIZE - (byte_offset & SOUND_BUFFER_SIZE_MASK);
		byte_count = sc;
		sc >>= 1;
	    }
	    if (bytes_so_far + byte_count > buffer_size)
	    {
		byte_count = buffer_size - bytes_so_far;
		if (byte_count == 0)
		    break;
		sc = byte_count;
		sc >>= 1;
	    }
//	    S9xMixSamplesO (Buf, sc, byte_offset & SOUND_BUFFER_SIZE_MASK);
	    S9xMixSamples (Buf + (byte_offset & SOUND_BUFFER_SIZE_MASK), sc);
	    so.samples_mixed_so_far += sc;
	    sample_count -= sc;
	    bytes_so_far = (so.samples_mixed_so_far << 1);
	    byte_offset += byte_count;
	} while (sample_count > 0);
    }
#endif
}


void update_sdl_audio(void *userdata, Uint8 * stream, int len)
{
#ifndef NOSOUND
    if (!snes4all_sound_enable)
	    return;
    int sample_count = len>>1;
    if (so.samples_mixed_so_far < sample_count)
    {
	int byte_offset = so.play_position + (so.samples_mixed_so_far << 1);
	S9xMixSamplesO (Buf, sample_count - so.samples_mixed_so_far,
			    byte_offset & SOUND_BUFFER_SIZE_MASK);
	so.samples_mixed_so_far = 0;
    }
    else
	so.samples_mixed_so_far -= sample_count;

 
//    if (!so.mute_sound)
#ifndef DREAMCAST
    memcpy(stream,Buf+so.play_position,len);
#else
    SDL_DC_SetSoundBuffer(Buf+so.play_position);
#endif
    so.play_position = (so.play_position + len) & SOUND_BUFFER_SIZE_MASK;
#endif
}


uint32 S9xReadJoypad (int which1)
{
	return joypad_val[which1];
}

#ifndef _ZAURUS
static int S9xCompareSDD1IndexEntries (const void *p1, const void *p2)
{
    return (*(uint32 *) p1 - *(uint32 *) p2);
}

void S9xLoadSDD1Data ()
{
    char filename [_MAX_PATH + 1];
    char index [_MAX_PATH + 1];
    char data [_MAX_PATH + 1];
    char patch [_MAX_PATH + 1];

    Memory.FreeSDD1Data ();

    strcpy (filename, S9xGetSnapshotDirectory ());

    if (strncmp (Memory.ROMName, "Star Ocean", 10) == 0)
	strcat (filename, "/socnsdd1");
    else
	strcat (filename, "/sfa2sdd1");

    DIR *dir = opendir (filename);

    index [0] = 0;
    data [0] = 0;
    patch [0] = 0;

    if (dir)
    {
	struct dirent *d;
	
	while ((d = readdir (dir)))
	{
	    if (strcasecmp (d->d_name, "SDD1GFX.IDX") == 0)
	    {
		strcpy (index, filename);
		strcat (index, "/");
		strcat (index, d->d_name);
	    }
	    else
	    if (strcasecmp (d->d_name, "SDD1GFX.DAT") == 0)
	    {
		strcpy (data, filename);
		strcat (data, "/");
		strcat (data, d->d_name);
	    }
	    if (strcasecmp (d->d_name, "SDD1GFX.PAT") == 0)
	    {
		strcpy (patch, filename);
		strcat (patch, "/");
		strcat (patch, d->d_name);
	    }
	}
	closedir (dir);

	if (strlen (index) && strlen (data))
	{
	    FILE *fs = fopen (index, "rb");
	    int len = 0;

	    if (fs)
	    {
		// Index is stored as a sequence of entries, each entry being
		// 12 bytes consisting of:
		// 4 byte key: (24bit address & 0xfffff * 16) | translated block
		// 4 byte ROM offset
		// 4 byte length
		fseek (fs, 0, SEEK_END);
		len = ftell (fs);
		rewind (fs);
		Memory.SDD1Index = (uint8 *) malloc (len);
		fread (Memory.SDD1Index, 1, len, fs);
		fclose (fs);
		Memory.SDD1Entries = len / 12;

		if (!(fs = fopen (data, "rb")))
		{
		    free ((char *) Memory.SDD1Index);
		    Memory.SDD1Index = NULL;
		    Memory.SDD1Entries = 0;
		}
		else
		{
		    fseek (fs, 0, SEEK_END);
		    len = ftell (fs);
		    rewind (fs);
		    Memory.SDD1Data = (uint8 *) malloc (len);
		    fread (Memory.SDD1Data, 1, len, fs);
		    fclose (fs);

		    if (strlen (patch) > 0 &&
			(fs = fopen (patch, "rb")))
		    {
			fclose (fs);
		    }
#ifdef MSB_FIRST
		    // Swap the byte order of the 32-bit value triplets on
		    // MSBFirst machines.
		    uint8 *ptr = Memory.SDD1Index;
		    for (int i = 0; i < Memory.SDD1Entries; i++, ptr += 12)
		    {
			SWAP_DWORD ((*(uint32 *) (ptr + 0)));
			SWAP_DWORD ((*(uint32 *) (ptr + 4)));
			SWAP_DWORD ((*(uint32 *) (ptr + 8)));
		    }
#endif
		    qsort (Memory.SDD1Index, Memory.SDD1Entries, 12,
			   S9xCompareSDD1IndexEntries);
		}
	    }
	}
	else
	{
	    printf ("Decompressed data pack not found in '%s'.\n", filename);
	}
    }
}
#endif
